package net.knarcraft.permissionsigns;

import net.knarcraft.permissionsigns.command.PermissionSignsCommand;
import net.knarcraft.permissionsigns.command.PermissionSignsTabCompleter;
import net.knarcraft.permissionsigns.container.PermissionSign;
import net.knarcraft.permissionsigns.container.SignCreationRequest;
import net.knarcraft.permissionsigns.formatting.Translator;
import net.knarcraft.permissionsigns.listener.SignListener;
import net.knarcraft.permissionsigns.manager.EconomyManager;
import net.knarcraft.permissionsigns.manager.PermissionManager;
import net.knarcraft.permissionsigns.manager.SignManager;
import net.knarcraft.permissionsigns.thread.PermissionTimeoutThread;
import net.knarcraft.permissionsigns.thread.SignCreationRequestTimeoutThread;
import net.milkbowl.vault.economy.Economy;
import net.milkbowl.vault.permission.Permission;
import org.bukkit.Bukkit;
import org.bukkit.command.PluginCommand;
import org.bukkit.entity.Player;
import org.bukkit.plugin.PluginDescriptionFile;
import org.bukkit.plugin.RegisteredServiceProvider;
import org.bukkit.plugin.ServicesManager;
import org.bukkit.plugin.java.JavaPlugin;
import org.bukkit.scheduler.BukkitScheduler;

import java.io.IOException;
import java.util.List;
import java.util.PriorityQueue;
import java.util.Queue;
import java.util.UUID;
import java.util.stream.Stream;

public final class PermissionSigns extends JavaPlugin {

    private static final Queue<SignCreationRequest> signCreationRequests = new PriorityQueue<>();
    private static String pluginVersion;
    private static PermissionSigns instance;

    /**
     * Instantiates the permission signs class
     */
    @SuppressWarnings("unused")
    public PermissionSigns() {
        super();
        instance = this;
    }

    /**
     * Gets an instance of this plugin
     *
     * @return <p>An instance of this plugin</p>
     */
    public static PermissionSigns getInstance() {
        return instance;
    }

    /**
     * Gets the version of this plugin
     *
     * @return <p>This plugin's version</p>
     */
    public static String getPluginVersion() {
        return pluginVersion;
    }

    /**
     * Adds a new sign creation request
     *
     * @param player <p>The player that initiated the sign creation</p>
     * @param sign   <p>The sign the player is about to create</p>
     */
    public static void addSignCreationRequest(Player player, PermissionSign sign) {
        signCreationRequests.add(new SignCreationRequest(sign, player, System.currentTimeMillis()));
    }

    /**
     * Gets the sign creation request for the player with the given UUID
     *
     * @param uuid <p>The UUID to get a sign creation request for</p>
     * @return <p>A sign creation request, or null if the UUID is not found</p>
     */
    public static SignCreationRequest getSignCreationRequest(UUID uuid) {
        Stream<SignCreationRequest> matchingRequests = signCreationRequests.stream().filter(
                (item) -> item.getPlayer().getUniqueId().equals(uuid));
        List<SignCreationRequest> requestList = matchingRequests.toList();
        if (!requestList.isEmpty()) {
            return requestList.get(0);
        } else {
            return null;
        }
    }

    /**
     * Cancels the sign creation request triggered by the user
     *
     * @param uuid <p>The UUID of the player to cancel the request for</p>
     */
    public static void cancelSignCreationRequest(UUID uuid) {
        Stream<SignCreationRequest> matchingRequests = signCreationRequests.stream().filter(
                (item) -> item.getPlayer().getUniqueId().equals(uuid));
        List<SignCreationRequest> requestList = matchingRequests.toList();
        if (requestList.size() > 0) {
            signCreationRequests.remove(requestList.get(0));
        }
    }

    @Override
    public void onEnable() {
        PluginDescriptionFile pluginDescriptionFile = this.getDescription();
        pluginVersion = pluginDescriptionFile.getVersion();

        // Plugin startup logic

        //TODO: Add commands create, add and remove
        // /ps create, /ps add, and /ps remove
        // create initiates the creation, and add adds properties
        // On creation, write "Creating PermissionSign", first asked for the name
        // Then asked for the permission. Allow several comma-separated permissions
        // Then asked for duration
        // Then asked for cost
        // Say "Sign completed! Right-click a sign to enable it."
        // Then asked to right-click a sign to create the new permission-sign
        // Perhaps ignore the old ways, and just have one command for creating permission signs: 
        // /ps create <name> <permission> <duration> <cost> to create a new permission-sign
        // Right-click a sign to create it
        // /ps cancel to cancel the sing creation
        // Break the sign to remove it, check for permission first
        // The name thing is probably useless, as the sign's location works as its id

        //TODO: Display and register the permission-sign
        // Start with [PermSign] in red
        // Next line is the permission node. Last child, upper-cased
        // Third line is n seconds
        // Last line is the cost, including the unit
        // Need to store any temporary permissions in a list/queue and have a thread which searches for expired 
        // permissions to de-register them

        //Not persistent, but might work as things shouldn't persist anyway
        //player.addAttachment(this, "essentials.fly", true, seconds * 20);
        //Vault probably has some API to add permissions

        //TODO: Store all temporary permissions with the time they were assigned and the time they should be granted.
        // Set the permissions as permanent and remove when the timer expires? Might cause problems for players that has
        // the permission inherited, unless we make sure the player does not have the permission before
        // Alternatively: Manage all permissions ourselves, for temporary permissions, and add permission attachments 
        // on startup I guess.

        //TODO: Start sign creation when the create command is used and save the data until an empty sign is right-clicked

        //TODO: Check for existence of old permission signs when clicked and register them as new permission signs. If 
        // it has the permissionSigns header and a name matching contents in signs.yml, add it.

        //Check if vault is loaded
        ServicesManager servicesManager = this.getServer().getServicesManager();
        RegisteredServiceProvider<Permission> permissionProvider = servicesManager.getRegistration(Permission.class);
        RegisteredServiceProvider<Economy> economyProvider = servicesManager.getRegistration(Economy.class);
        if (permissionProvider != null && economyProvider != null) {
            EconomyManager.initialize(economyProvider.getProvider());
            PermissionManager.initialize(permissionProvider.getProvider());
        } else {
            throw new IllegalStateException("[PermissionSigns] Error: Vault could not be loaded");
        }
        getServer().getPluginManager().registerEvents(new SignListener(), this);

        Translator.loadLanguages("en");
        registerCommands();
        BukkitScheduler scheduler = Bukkit.getScheduler();
        scheduler.runTaskTimer(this, new SignCreationRequestTimeoutThread(signCreationRequests), 0L, 100L);
        scheduler.runTaskTimer(this, new PermissionTimeoutThread(), 0L, 25L);
        SignManager.loadSigns();
    }

    /**
     * Registers a command for this plugin
     */
    private void registerCommands() {
        PluginCommand stargateCommand = this.getCommand("permissionsigns");
        if (stargateCommand != null) {
            stargateCommand.setExecutor(new PermissionSignsCommand());
            stargateCommand.setTabCompleter(new PermissionSignsTabCompleter());
        }
    }

    @Override
    public void onDisable() {
        try {
            SignManager.saveSigns();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

}
